/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file logging.c
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 * 
 * \brief Implementation of the event logger.
 */

#include <stdio.h>
#include <sys/event.h>
#include <sys/heap.h>
#include <sys/timer.h>
#include <sys/mutex.h>
#include <string.h>
#include <debug/xrlog.h>
#include <debug/logging.h>
#include <sys/tracer.h>

#define LOG_BANK_START   			(129L * 1024)
#define LOG_BANK_SIZE				(10 * 1024)

#define LINEBUFFER_SIZE				255

#define LOG_MAX_LISTENERS			3

#define LOG_MSG_HDR_SIZE			(sizeof(log_msg_t) - 1)

#define LOG_MSG_TAIL_SIZE			(sizeof(log_msg_tail))

// returns true if the log level filter for the level passed is on.
#define LOG_FILTER_ON(filter, level) 		(filter & (1 << (level - 1)))

#define MIN(a, b) ((a > b)? b : a)

// trace setup
#define LOG_TRACE_CREATE_TSTAMP		0
#define LOG_TRACE_CREATE_MSG		1
#define LOG_TRACE_WRITE				2
#define LOG_TRACE_DONE				3
                                          
/* already defined in hardware/ram_banks.h */
//#define MIN(a,b) ((a>b)?b:a)

typedef struct _log_listener_s {
	u_char* log_filter;
	LOG_DATA_CB;
	void* cb_arg;
} _log_listener;

typedef struct _log_stack {
    xrlog_t* log_buf;
    u_short log_hdl;
    log_filter_t logmask;
    char write_buf[LINEBUFFER_SIZE + 1 + LOG_MSG_HDR_SIZE];		// + 1 for termination string
    char read_buf[LINEBUFFER_SIZE + 1 + LOG_MSG_HDR_SIZE];
    int linesize;
    MUTEX read_buf_mutex;
    HANDLE line_added;
    _log_listener listeners[LOG_MAX_LISTENERS];
    u_char nr_listeners;
} _log_stack_t;

static _log_stack_t log_stack;

extern u_int micros_high;

extern int _putf(int _putb(int, CONST void *, size_t), int fd, CONST char *fmt, va_list ap);


static int _sputb(int fd, CONST void *buffer, size_t count)
{
    char **spp = (char **) ((uptr_t) fd);
    int copy_len;
    copy_len = MIN( count, LINEBUFFER_SIZE - log_stack.linesize);
    memcpy(*spp, buffer, copy_len);
    *spp += copy_len;
    log_stack.linesize += copy_len;
    return copy_len;
}

int _log_msg_to_str(u_char* buf, CONST char* fmt, va_list ap)
{
	log_stack.linesize = 0;
	// print message to buffer passed
	_putf(_sputb, (int) ((uptr_t)&buf), fmt, ap);
	// make shure that there is a newline if the string is cutted
    if (log_stack.linesize == LINEBUFFER_SIZE)
    	*(buf - 1) = '\n';
	// add null char
	*buf++ = 0;
	return log_stack.linesize + 1;
}

int log_msg_extract(log_msg_t* msg)
{
	log_msg_args_t args;
	log_msg_tail* tail = (log_msg_tail*)(msg->data + msg->len);
	char* fmt = tail->fmt;
	
	// only extract if the message is packed (msg->fmt != NULL)
	if (fmt) {
		// copy message args
		memcpy(args, msg->data, sizeof(log_msg_args_t));
		// build output string
		msg->len = _log_msg_to_str(msg->data, fmt, (va_list)args);
	}
	// return size of message
	return (msg->len);
}

void log_write(u_char class, u_char level, CONST char* fmt, ...)
{
	va_list ap;
	log_msg_t* msg;
	char* buf;
	u_short msg_len;
	log_msg_tail* tail;
	
	msg = (log_msg_t*)log_stack.write_buf;
	buf = msg->data;
	
    if(LOG_FILTER_ON(log_stack.logmask[class], level)) {
		
//		TRACE_ADD_ITEM(TRACE_TAG_USER, LOG_TRACE_CREATE_TSTAMP);
		
		/* get time stamp */
		msg->t_stamp = NutGetMillis();
		
//		TRACE_ADD_ITEM(TRACE_TAG_USER, LOG_TRACE_CREATE_MSG);
		
		/* build msg string */
		va_start(ap, fmt);
		msg_len = _log_msg_to_str(msg->data, fmt, ap);
		va_end(ap);
	    
	    /* build msg header */
		msg->class = class;
		msg->level = level;
		msg->len = msg_len;
	    
	    /* build msg tail */
	    tail = (log_msg_tail*)(msg->data + msg_len);
	    tail->fmt = NULL;
	    tail->msg_len = msg_len;
	    
//	    TRACE_ADD_ITEM(TRACE_TAG_USER, LOG_TRACE_WRITE);
	    
	    // write message
	    xrlog_write(log_stack.log_buf, (u_char*)msg, LOG_MSG_HDR_SIZE + msg_len + LOG_MSG_TAIL_SIZE);
	    
//	    TRACE_ADD_ITEM(TRACE_TAG_USER, LOG_TRACE_DONE);
    }
    
    /* wake up event listener */
    NutEventBroadcastAsync(&log_stack.line_added);
}

void log_add_msg(u_char class, u_char level, CONST char* fmt, ...)
{
	u_char i;
    va_list ap;
	log_msg_t* msg;
	unsigned int* msg_args;
	log_msg_tail* tail;
	
	msg = (log_msg_t*)log_stack.write_buf;
	
    if(LOG_FILTER_ON(log_stack.logmask[class], level)) {
    	
//    	TRACE_ADD_ITEM(TRACE_TAG_USER, LOG_TRACE_ADD_MSG);
    	
		// build message header
		msg->t_stamp = NutGetMillis();
	    msg->class = class;
	    msg->level = level;
	    msg->len = sizeof(log_msg_args_t);
	    
	    // store message args
		va_start(ap, fmt);
		msg_args = (unsigned int*)msg->data;
		for (i=0; i<LOG_MAX_NR_ARGS; i++) {
			msg_args[i] = va_arg(ap, int);
		}
		va_end(ap);
		
		// build msg tail
		tail = (log_msg_tail*)(msg->data + msg->len);
		tail->fmt = (char*)fmt;
		tail->msg_len = msg->len;
		
		// write msg
	    xrlog_write(log_stack.log_buf, (u_char*)msg, LOG_MSG_HDR_SIZE + msg->len + LOG_MSG_TAIL_SIZE);
	    
//	    TRACE_ADD_ITEM(TRACE_TAG_USER, LOG_TRACE_DONE);
    }
    
    /* inform listeners */
    NutEventPostAsync(&log_stack.line_added);
}

inline void log_empty(void)
{
	// empty the log buffer
	xrlog_empty(log_stack.log_buf);
	// reset log handle of verbosity thread
	log_stack.log_hdl = xrlog_get_curr_pos(log_stack.log_buf);
}

void log_get_messages(u_short nr_msgs, u_char class, u_char mask, LOG_DATA_CB, void* cb_arg)
{
	log_msg_tail tail;
	log_msg_t* msg = (log_msg_t*)log_stack.read_buf;
	u_short hdl;
	u_short count = 0;
	
	if (class > LOG_MAX_CLASSES) {
		class = 0;
	}
	
	// lock the read buffer
	NutMutexLock(&log_stack.read_buf_mutex);
	
	// get buffer pointer
	hdl = xrlog_get_curr_pos(log_stack.log_buf);
	
	// seek back N messages
	while ((nr_msgs == LOG_ALL_MESSAGES) || (count != nr_msgs)) {
		// read next message tail
		if (xrlog_seek(log_stack.log_buf, &hdl, -((short)LOG_MSG_TAIL_SIZE)))
			break;
		xrlog_read(log_stack.log_buf, (u_char*)&tail, &hdl, LOG_MSG_TAIL_SIZE);
		
		// seek to next message - stop if end of data is reached
		if (xrlog_seek(log_stack.log_buf, &hdl, -(LOG_MSG_HDR_SIZE + tail.msg_len + LOG_MSG_TAIL_SIZE)))
			break;
		
		/*
		 * increment count iff class & level matches.
		 * 
		 * Note: we do not check if the end of the log buf has been reached.
		 * This should be ok because the reading and back-seeking is atomic
		 */
		// read header
		xrlog_read(log_stack.log_buf, (u_char*)msg, &hdl, LOG_MSG_HDR_SIZE);
		// if class and level matches, increment count
		if(((class == LOG_ALL_CLASSES) || (msg->class == class)) && (LOG_FILTER_ON(mask, msg->level))) {
			count++;
		}
		// seek back to the end of the next message
		xrlog_seek(log_stack.log_buf, &hdl, -((short)LOG_MSG_HDR_SIZE));
		
		// give other threads a chance!
		NutThreadYield();
	}
	
	// print out maximal 'count' messages
	while (count) {
		// try to read next msg header
		if (xrlog_read(log_stack.log_buf, (u_char*)msg, &hdl, LOG_MSG_HDR_SIZE))
			break;
			
		// only read message if level and class matches
		if(((class == LOG_ALL_CLASSES) || (msg->class == class)) && (LOG_FILTER_ON(mask, msg->level))) {
			// read msg & msg tail
			if (xrlog_read(log_stack.log_buf, (u_char*)msg->data, &hdl, msg->len + LOG_MSG_TAIL_SIZE))
				break;
			// print msg, i.e. call user callback
			log_data_cb(msg, cb_arg);
			// decrement msg count
			count--;
		}
		// else, just seek
		else {
			if (xrlog_seek(log_stack.log_buf, &hdl, msg->len + LOG_MSG_TAIL_SIZE))
				break;
		}
		
		// give other threads a chance!
		NutThreadYield();
	}
	
	// unlock read buffer
	NutMutexUnlock(&log_stack.read_buf_mutex);
}


THREAD(verbose_log, arg)
{
	u_short i;
	log_msg_t* msg;
	
	msg = (log_msg_t*)log_stack.read_buf;
	_log_listener* listeners = log_stack.listeners;
	
	// run this service in background
	NutThreadSetPriority(200);
	
	while(1) {
		// wait 'til a new line has been added
		NutEventWait(&log_stack.line_added, NUT_WAIT_INFINITE);
		// get access to the read buffer
		NutMutexLock(&log_stack.read_buf_mutex);
		
		/* print out new log messages */
		while (1) {
			// read next message header
			if (xrlog_read(log_stack.log_buf, (u_char*)msg, &log_stack.log_hdl, LOG_MSG_HDR_SIZE))
				break;
				
			// read rest of the message (including tail)
			if (xrlog_read(log_stack.log_buf, (u_char*)msg->data, &log_stack.log_hdl, msg->len + LOG_MSG_TAIL_SIZE))
				break;
			
			// signal callback of registered listeners
			for (i=0; i<LOG_MAX_LISTENERS; i++) {
			    if(listeners[i].log_filter && LOG_FILTER_ON(listeners[i].log_filter[msg->class], msg->level)) {
			    	// call log data callback
					listeners[i].log_data_cb(msg, listeners[i].cb_arg);
			    }
			}
			
			// give other threads a chance!
			NutThreadYield();
		}
		
		// release lock on read buffer
		NutMutexUnlock(&log_stack.read_buf_mutex);
	}
}

short log_register_event_listener(LOG_DATA_CB, void* cb_arg, log_filter_t filter)
{
	u_char i;
	_log_listener* listener = log_stack.listeners;
	
	// try to get an empty entry in the list of listeners
	for (i=0; i<LOG_MAX_LISTENERS; i++) {
		if (!listener[i].log_filter) break;
		listener++;
	}
	// return error if list is full
	if (i == LOG_MAX_LISTENERS) return -1;
	
	// else, register listener
	listener = &log_stack.listeners[i];
	listener->log_filter = filter;
	listener->log_data_cb = log_data_cb;
	listener->cb_arg = cb_arg;
	return 0;
}


void log_set_logmask( u_char class, u_char mask)
{
    if( class > 0 && class < LOG_MAX_CLASSES) {
        log_stack.logmask[class] = mask;
    }
}

short log_get_logmask(u_char class)
{
	// get desired mask and return it
	if (class > 0 && class < LOG_MAX_CLASSES)
		return log_stack.logmask[class];
	// return error if class not valid
	return -1;
}
		
void log_init(void)
{
	u_short i;
    
    // init xrlog
    log_stack.log_buf = xrlog_init(LOG_BANK_START, LOG_BANK_SIZE);

	// init read buffer mutex
    NutMutexInit( &log_stack.read_buf_mutex);

	// init logmasks
	for (i=0; i<LOG_MAX_CLASSES; i++) {
		log_stack.logmask[i] = LOG_MASK_ALL;
	}
	
	// init log verbosity thread & log handle it should use
	log_stack.log_hdl = xrlog_get_curr_pos(log_stack.log_buf);
	NutThreadCreate("log", verbose_log, 0, 1024);
	
//	// trace setup
//	for(i = TRACE_TAG_FIRST; i <= TRACE_TAG_LAST; i++)
//      NutTraceMaskClear(i);
//
//	NutTraceRegisterUserTag(LOG_TRACE_CREATE_TSTAMP, "creating tstamp... ");
//  NutTraceRegisterUserTag(LOG_TRACE_CREATE_MSG, "building log msg... ");
//  NutTraceRegisterUserTag(LOG_TRACE_WRITE, "write to log... ");
//  NutTraceRegisterUserTag(LOG_TRACE_DONE, "done.");
//    
//  NutTraceMaskSet(TRACE_TAG_USER);
//  NutTraceInit(1000, TRACE_MODE_CIRCULAR);
}
