/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file bt_acl_defs.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 * 
 * \brief API for using the ACL layer.
 * 
 * API for using the ACL layer.
 * 
 * Contains:
 * - Definitions of maximal ACL packet payload sizes
 * - Definitions of the ACL packet buffers used on the receiving side
 * to buffer (connection-less) ACL data packets (used by the l2cap
 * connection-less layer)
 * - API for creating / destroying ACL packet buffers
 * - API for sending ACL data
 * - API for accessing ACL packet information
 */
#ifndef _BT_ACL_DEFS_H_
#define _BT_ACL_DEFS_H_

#include <bt/bt_hci_defs.h>
#include <bt/bt_semaphore.h>

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DM1.
 */
#define BT_ACL_PAYLOAD_SIZE_DM1			16

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DH1.
 */
#define BT_ACL_PAYLOAD_SIZE_DH1			26

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DM3.
 */
#define BT_ACL_PAYLOAD_SIZE_DM3			121

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DH3.
 */
#define BT_ACL_PAYLOAD_SIZE_DH3			183

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DM5.
 */
#define BT_ACL_PAYLOAD_SIZE_DM5			224

/**
 * Maximal amount of user data (in bytes) that can be sent over an ACL in
 * a single packet using packets of type DH5.
 */
#define BT_ACL_PAYLOAD_SIZE_DH5			339

/**
 * Size of the HCI packet header that gets attached to each acl packet
 * 
 * Note: includes the hci-type field.
 */
#define BT_ACL_HCI_PKT_HDR_LEN		5

/**
 * \brief Definition of an ACL communication link.
 * 
 * Variables of type bt_hci_acl_link serve as a handle for sending ACL
 * packets to a remote device using an exisisting baseband connection.
 * 
 * Get a handle using #bt_acl_link_open, write your data using
 * #bt_acl_link_write, and free the link using #bt_acl_link_close.
 */
struct bt_hci_acl_link {
	bt_hci_con_handle_t con_hdl;
	u_short data_len;
	u_short written;
	struct btstack* stack;
	struct bt_semaphore sema;
};

/**
 * \brief Definition of an ACL packet buffer.
 * 
 * ACL packet buffers are used on the receiving side to buffer
 * (connection-less) data packets. Each ACL packet buffer is able to buffer
 * a single ACL packet.
 * 
 * Create a single ACL packet buffer using #bt_acl_pkt_buf_create, create
 * multiple ACL packet buffers using #bt_acl_pkt_queue_create.
 * 
 * Destroy a single ACL packet buffer by calling #bt_acl_pkt_buf_destroy,
 * and destroy an ACL packet buffer queue using #bt_acl_pkt_queue_destroy.
 */
typedef struct bt_acl_pkt_buf_s {
	struct bt_hci_pkt_acl* pkt;
	struct bt_acl_pkt_buf_s* next;
	struct bt_acl_pkt_buf_s* next_frag;
	u_char* pload;
	u_short plen;
	u_long tstamp;
	void* arg;
} bt_acl_pkt_buf;

/**
 * \brief Initializes the ACL layer.
 * 
 * Sets the ACL packet type(s) to use and creates the first ACL buffer.
 * 
 * Note: When dealing with ACL data packets, this function should be called
 * immediately after the bt-stack has been initialized! The reason for this
 * is that setting the ACL packet types to use, determines the size of
 * the ACL buffers to be used on the receiving side.
 * 
 * \param stack The bt-stack to use
 * \param pkt_types ACL packet type(s) to use
 * 
 * \return Maximal ACL pkt payload size, corresponding to the packet types
 * chosen.
 * 
 * #see bt_acl_pkt_buf
 * #see bt_acl_pkt_buf_create
 * #see bt_hci_def.h
 */
u_short bt_acl_init(struct btstack* stack, u_short pkt_types);

/**
 * \brief Initializes a given baseband connection to send ACL data.
 * 
 * Basically, this function provides unique access to the bt uart and
 * writes the information needed by the hci layer to send acl data over
 * a given baseband communication link to the bt controller.
 * 
 * @param stack The bt-stack to use
 * @param con_handle Connection handle of the desired baseband connection
 * @param pb_flag Packet Boundary Flag #HCI_PB_FIRST_FRAGMENT or #HCI_PB_CONTD_FRAGMENT
 * @param bc_flag Broadcast Flag #HCI_BC_UNICAST, #HCI_BC_ACTIVE or #HCI_BC_PICONET
 * @param length Amount of ACL data to send
 * @param acl_link Pointer to a variable where the pointer to the initiated
 * acl communication link shall be written to. Use the pointer to the acl link
 * obtained to write your data to the link (see #bt_hci_acl_link_write).
 * 
 * Note: ACL data can be sent to the con handle #BT_HCI_CON_HANDLE_OWN, i.e.
 * to the sending device itself. Quite useful for testing protocols...
 * 
 * \see #bt_hci_acl_link_write()
 * \see #bt_hci_acl_link_close()
 */
long bt_acl_link_open(struct btstack* stack,
						bt_hci_con_handle_t con_handle,
						u_char pb_flag,
						u_char bc_flag,
						u_short length,
						struct bt_hci_acl_link** acl_link);

/**
 * @brief Writes ACL data to the baseband connection given.
 * 
 * Use #bt_hci_acl_link_open() to get access to the desired baseband connection.
 * Write your data, and release the connection using
 * #bt_hci_acl_link_close().
 * 
 * @param acl_link Pointer to the ACL communication link
 * (see #bt_hci_uart_lock())
 * @param data Pointer to the data to be written
 * @param len Number of bytes to be written
 * 
 * @see #bt_hci_acl_link_open()
 * @see #bt_hci_acl_link_close()
 */
inline long bt_acl_link_write(struct bt_hci_acl_link* acl_link, u_char* data, u_short len);

/**
 * @brief Releases a given baseband connection.
 * 
 * Call this function after you've written your data to the bt-controller
 * using #bt_hci_acl_link_write.
 * 
 * @param acl_link Pointer to the previousely initiated acl communication link
 * (see #bt_hci_acl_link_open).
 * 
 * @see #bt_hci_acl_link_open()
 * @see #bt_hci_acl_link_write()
 */
inline void bt_acl_link_close(struct bt_hci_acl_link* acl_link);

/**
 * \brief Creates a single ACL packet buffer.
 * 
 * Creates a single ACL packet buffer. Destroy the buffer created using
 * #bt_acl_pkt_buf_destroy.
 * 
 * Note: bt_acl_init() must have been called before calling this function.
 * 
 * \returns
 * - Pointer to the created buffer
 * - 0 if not enough memory available
 * 
 * \see #bt_acl_init
 * \see #bt_acl_pkt_buf
 * \see #bt_acl_pkt_buf_destroy
 */
bt_acl_pkt_buf* bt_acl_pkt_buf_create(struct btstack* stack);

/**
 * \brief Destroys the ACL packet buffer passed.
 * 
 * Destroys the previousely created (using #bt_acl_pkt_buf_create)
 * ACL packet buffer.
 * 
 * \see #bt_acl_pkt_buf
 * \see #bt_acl_pkt_buf_create
 */
void bt_acl_pkt_buf_destroy(bt_acl_pkt_buf* buf);

/**
 * \brief Creates an ACL packet buffer queue
 * 
 * Creates a queue of ACL packet buffers of the size given. Destroy the queue
 * created using #bt_acl_pkt_queue_destroy.
 * 
 * Note: bt_acl_init() must have been called before calling this function.
 * 
 * \param nr_bufs Number of buffers to create
 * 
 * \returns
 * - Pointer to the created packet queue
 * - 0 if not enough memory available
 * 
 * \see #bt_acl_pkt_buf
 * \see #bt_acl_pkt_queue_destroy
 */
bt_acl_pkt_buf* bt_acl_pkt_queue_create(struct btstack* stack, u_short nr_bufs);

/**
 * \brief Destroys an ACL packet buffer queue
 * 
 * Destroys the previousely created (using #bt_acl_pkt_queue_create) queue
 * of ACL packet buffers.
 * 
 * \param que Pointer to the first packet buf of the queue
 * 
 * \see #bt_acl_pkt_buf
 * \see #bt_acl_pkt_queue_create
 */
void bt_acl_pkt_queue_destroy(bt_acl_pkt_buf* que);

/**
 * Returns setting of the packet types to use.
 */
inline u_short bt_acl_get_pkt_types(struct btstack* stack);

/**
 * Returns maximal size of ACL packet payload.
 */
inline u_short bt_acl_get_max_pkt_len(struct btstack* stack);

/**
 * Returns a pointer to the payload of an ACL packet
 */
inline u_char* bt_acl_get_data_pointer(struct bt_hci_pkt_acl* pkt);

/**
 * Returns the con handle the packet passed was received on.
 */
inline bt_hci_con_handle_t bt_acl_get_con_handle(struct bt_hci_pkt_acl* pkt);

/**
 * Returns the value of the fragmentation flag of the packet passed.
 */
inline u_char bt_acl_get_pb_flag(struct bt_hci_pkt_acl* pkt);

/**
 * Returns the value of the broadcast flag of the packet passed.
 */
inline u_char bt_acl_get_bc_flag(struct bt_hci_pkt_acl* pkt);

#endif //_BT_ACL_DEFS_H_
