/*
 * Copyright (C) 2000-2006 by ETH Zurich
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY ETH ZURICH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ETH ZURICH
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.btnode.ethz.ch
 *
 */

/**
 * \file bt_msg_defs.h
 * 
 * \author Kevin Martin <kevmarti@tik.ee.ethz.ch> 
 * 
 * \date 2006/03/23
 */
#ifndef _BT_MSG_DEFS_H_
#define _BT_MSG_DEFS_H_

#include <sys/types.h>
#include <bt/bt_acl_defs.h>

/**
 * Definition of a bt message.
 * 
 * A bt message consists of several #bt_acl_pkt_buf's (fragments) tied
 * together (using the #bt_acl_pkt_buf::next_frag pointers).
 * 
 * Note: This definition was intended to distinguish between whole messages
 * and single #bt_acl_pkt_buf's in case of packet fragmentation / reassembly.
 * But up to now, packet fragmentation has not been implemented on the
 * connection-less side (l2cap does fragmentation, but uses a different
 * scheme).
 */
typedef bt_acl_pkt_buf bt_msg;

typedef struct bt_msg_fifo_s {
	bt_msg* first;
	bt_msg* last;
	u_short size;
} bt_msg_fifo;

/**
 * \brief Creates a message fifo queue.
 * 
 * Creates a fifo queue for queueing bt messages (see #bt_msg). Destroy the
 * queue created using #bt_msg_fifo_destroy.
 * 
 * Note: #bt_acl_init must have been called before calling this function.
 * 
 * \param nr_bufs Number of buffers to create
 * 
 * \returns
 * - Pointer to the created packet queue
 * - 0 if not enough memory available
 * 
 * \see #bt_msg
 * \see #bt_msg_fifo_destroy
 */
bt_msg_fifo* bt_msg_fifo_create(void);

/**
 * \brief Destroys a bt message fifo queue
 * 
 * Destroys the previousely created (using #bt_msg_fifo_create) 
 * message fifo queue.
 * 
 * \param que Pointer to the fifo queue
 * 
 * \see #bt_msg_fifo_create
 */
void bt_msg_fifo_destroy(bt_msg_fifo* fifo);

/**
 * \brief Gets the message most recently queued.
 */
bt_msg* bt_msg_fifo_get_first(bt_msg_fifo* fifo);

/**
 * \brief Adds the message passed to the head of a message fifo queue.
 */
void bt_msg_fifo_add_first(bt_msg_fifo* fifo, bt_msg* msg);

/**
 * \brief Adds the message passed to the tail of a message fifo queue.
 */
void bt_msg_fifo_add_last(bt_msg_fifo* fifo, bt_msg* msg);

#endif //_BT_MSG_DEFS_H_
